package ij2x.plugin.event;


/**
 *
 * @author Mahdi Rawak
 */
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsEnvironment;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;

import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;


import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.WeakHashMap;

import javax.swing.JComponent;
import javax.swing.JRootPane;
import javax.swing.JWindow;

import javax.swing.plaf.basic.BasicRootPaneUI;

import javax.swing.event.MouseInputListener;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;


public class MouseInputHandler implements MouseInputListener, ComponentListener  {
    private JComponent target;
    private Point start_drag;
    private Point start_loc;
    private boolean dragging;
    private long mouseDownTime;
    public static final int DOUBLE_CLICK_THRESHOLD = 150;
    private JRootPane root;
    /**
     * Window the <code>JRootPane</code> is in.
     */
    private Window window;
    /**
     * <code>JComponent</code> providing window decorations. This will be
     * null if not providing window decorations.
     */
    private JComponent titlePane;
    /**
     * Since method Window.getWindows() is only available since Java 1.6,
     * we indirectly keep track of all windows by ourselves by storing
     * all JRootPanes in this weak hash map.
     * We add a JRootPane to this map, upon installUI, and remove a JRootPane
     * from this upoin deinstallUI.
     */
    private static WeakHashMap allRootPanes = new WeakHashMap();
    /**
    /**
     * <code>Cursor</code> used to track the cursor set by the user.
     * This is initially <code>Cursor.DEFAULT_CURSOR</code>.
     */
    private Cursor lastCursor =
            Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);

    /**
     * Height and width of resize handle on the lower right corner of the window.
     * FIXME - This value depends on font size.
     */
    private static final int BORDER_DRAG_THICKNESS = 15;

    /**
     * Set to true if the drag operation is moving the window.
     */
    private boolean isMovingWindow;
    /**
     * Used to determine the corner the resize is occuring from.
     */
    private int dragCursor;
    /**
     * X location the mouse went down on for a drag operation.
     */
    private int dragOffsetX;
    /**
     * Y location the mouse went down on for a drag operation.
     */
    private int dragOffsetY;
    /**
     * Width of the window when the drag started.
     */
    private int dragWidth;
    /**
     * Height of the window when the drag started.
     */
    private int dragHeight;
    /**
     * We cache the screen bounds here, so that we don't have to retrieve
     * them for each mouseDragged event. We clear the cache on mouseReleased.
     */
    private Rectangle cachedScreenBounds;

    public void mousePressed(MouseEvent ev) {
        JRootPane rootPane = getRootPane();

        if (rootPane.getWindowDecorationStyle() == JRootPane.NONE) {
            return;
        }
        Point dragWindowOffset = ev.getPoint();
        Window w = (Window) ev.getSource();
        if (w != null) {
            w.toFront();
        }
        Point convertedDragWindowOffset = SwingUtilities.convertPoint(
                w, dragWindowOffset, getTitlePane());

        Frame f = null;
        Dialog d = null;

        if (w instanceof Frame) {
            f = (Frame) w;
        } else if (w instanceof Dialog) {
            d = (Dialog) w;
        }

        int frameState = (f != null) ? f.getExtendedState() : 0;

        if (getTitlePane() != null &&
                getTitlePane().contains(convertedDragWindowOffset)) {
            if (f != null && ((frameState & Frame.MAXIMIZED_BOTH) == 0) || (d != null)) {
                isMovingWindow = true;
                dragOffsetX = dragWindowOffset.x;
                dragOffsetY = dragWindowOffset.y;
            }
        } else if (f != null && f.isResizable() && ((frameState & Frame.MAXIMIZED_BOTH) == 0) || (d != null && d.isResizable())) {
            dragOffsetX = dragWindowOffset.x;
            dragOffsetY = dragWindowOffset.y;
            dragWidth = w.getWidth();
            dragHeight = w.getHeight();
            dragCursor =
                    (dragWindowOffset.x >= dragWidth - BORDER_DRAG_THICKNESS &&
                    dragWindowOffset.y >= dragHeight - BORDER_DRAG_THICKNESS) ? Cursor.SE_RESIZE_CURSOR : 0;
        }
    }

    public void mouseReleased(MouseEvent ev) {
        if (dragCursor != 0 && window != null && !window.isValid()) {
            // Some Window systems validate as you resize, others won't,
            // thus the check for validity before repainting.
            window.validate();
            getRootPane().repaint();
        }
        isMovingWindow = false;
        dragCursor = 0;
        cachedScreenBounds = null;
    }

    public void mouseMoved(MouseEvent ev) {
    }

    private void adjust(Rectangle bounds, Dimension min, int deltaX,
            int deltaY, int deltaWidth, int deltaHeight) {
        bounds.x += deltaX;
        bounds.y += deltaY;
        bounds.width += deltaWidth;
        bounds.height += deltaHeight;
        if (min != null) {
            if (bounds.width < min.width) {
                int correction = min.width - bounds.width;
                if (deltaX != 0) {
                    bounds.x -= correction;
                }
                bounds.width = min.width;
            }
            if (bounds.height < min.height) {
                int correction = min.height - bounds.height;
                if (deltaY != 0) {
                    bounds.y -= correction;
                }
                bounds.height = min.height;
            }
        }
    }

    public void mouseDragged(MouseEvent ev) {
        Window w = (Window) ev.getSource();
        Point pt = ev.getPoint();

        if (isMovingWindow) {
            // Sometimes we get mouse dragged events even when we are not
            // showing on screen (?)
            if (w.isShowing()) {
                Point windowPt = w.getLocationOnScreen();

                windowPt.x += pt.x - dragOffsetX;
                windowPt.y += pt.y - dragOffsetY;

                boolean isOnDefaultScreen =
                        w.getGraphicsConfiguration().getDevice() ==
                        GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();

                // If an edge of the window is within the snap distance of the
                // edge of another window, then align it to it.
                // ----------------------------------------------------------
                int snap = UIManager.getInt("RootPane.windowSnapDistance");
                if (snap > 0 && (ev.getModifiersEx() & InputEvent.ALT_DOWN_MASK) == 0) {
                    // Collects all bounds to which we want to snap to
                    LinkedList snapBounds;
                    // Collect window bounds
                    do {
                        snapBounds = new LinkedList();
                        try {
                            for (Iterator i = allRootPanes.keySet().iterator(); i.hasNext();) {
                                JRootPane otherRootPane = (JRootPane) i.next();
                                Window other = SwingUtilities.getWindowAncestor(otherRootPane);
                                if (other != null && other.isShowing() && other != w) {
                                    snapBounds.add(other.getBounds());
                                }
                            }
                        } catch (ConcurrentModificationException e) {
                            // allRootPanes is a WeakHashMap, thus iterating over
                            // it may fail sometimes because the garbage collector
                            // removes items in a worker thread.
                            snapBounds = null;
                        }
                    } while (snapBounds == null);

                    // Collect screen bounds
                    snapBounds.add(w.getGraphicsConfiguration().getBounds());
                    if (isOnDefaultScreen) {
                        Rectangle r = w.getGraphicsConfiguration().getBounds();
                        Insets insets = w.getToolkit().getScreenInsets(w.getGraphicsConfiguration());
                        r.x += insets.left;
                        r.y += insets.top;
                        r.width -= insets.left + insets.right;
                        r.height -= insets.top + insets.bottom;
                        snapBounds.add(r);
                    }

                    Dimension windowDim = w.getSize();
                    Rectangle windowRect = new Rectangle(windowPt.x, windowPt.y, windowDim.width, windowDim.height);
                    Rectangle snapper = new Rectangle();
                    for (Iterator i = snapBounds.iterator(); i.hasNext();) {
                        Rectangle r = (Rectangle) i.next();

                        snapper.setBounds(r);
                        snapper.grow(snap, snap);
                        if (snapper.intersects(windowRect)) {

                            if (windowPt.x > r.x - snap &&
                                    windowPt.x < r.x + snap) {
                                // align my left edge to frame left edge
                                windowPt.x = r.x;
                            } else if (windowPt.x > r.x + r.width - snap &&
                                    windowPt.x < r.x + r.width + snap) {
                                // align my left edge to frame right edge
                                windowPt.x = r.x + r.width;
                            } else if (windowPt.x + windowDim.width > r.x - snap &&
                                    windowPt.x + windowDim.width < r.x + snap) {
                                // align my right edge to frame left edge
                                windowPt.x = r.x - windowDim.width;
                            } else if (windowPt.x + windowDim.width > r.x + r.width - snap &&
                                    windowPt.x + windowDim.width < r.x + r.width + snap) {
                                // align my right edge to frame right edge
                                windowPt.x = r.x + r.width - windowDim.width;
                            }
                            if (windowPt.y > r.y - snap && windowPt.y < r.y + snap) {
                                // align my top edge to frame top edge
                                windowPt.y = r.y;
                            } else if (windowPt.y > r.y + r.height - snap &&
                                    windowPt.y < r.y + r.height + snap) {
                                // align my top edge to frame bottom edge
                                windowPt.y = r.y + r.height;
                            } else if (windowPt.y + windowDim.height > r.y - snap &&
                                    windowPt.y + windowDim.height < r.y + snap) {
                                // align my bottom edge to frame top edge
                                windowPt.y = r.y - windowDim.height;
                            } else if (windowPt.y + windowDim.height > r.y + r.height - snap &&
                                    windowPt.y + windowDim.height < r.y + r.height + snap) {
                                // align my bottom edge to frame bottom edge
                                windowPt.y = r.y + r.height - windowDim.height;
                            }
                        }
                    }
                }
                // Constrain windowPt in order to ensure that a portion of the
                // title pane is always visible on screen
                // ----------------------------------------------------------
                // Get usable screen bounds
                if (isOnDefaultScreen) {
                    if (cachedScreenBounds == null) {
                        cachedScreenBounds = w.getGraphicsConfiguration().getBounds();
                        Insets screenInsets = w.getToolkit().getScreenInsets(w.getGraphicsConfiguration());
                        cachedScreenBounds.x += screenInsets.left;
                        cachedScreenBounds.y += screenInsets.top;
                        cachedScreenBounds.width -= screenInsets.left + screenInsets.right;
                        cachedScreenBounds.height -= screenInsets.top + screenInsets.bottom;
                    }
                    Rectangle titlePaneBounds = getTitlePane().getBounds();
                    Dimension windowSize = window.getSize();

                    if (isVertical(getRootPane())) {
                        // For vertical title bar, title pane must be fully visible
                        // on x-axis, and at least 20 pixel on y-axis.
                        windowPt.x = Math.max(cachedScreenBounds.x + titlePaneBounds.x, windowPt.x);
                        windowPt.x = Math.min(cachedScreenBounds.x + cachedScreenBounds.width -
                                titlePaneBounds.x - titlePaneBounds.width, windowPt.x);

                        windowPt.y = Math.max(cachedScreenBounds.y - windowSize.height + 20, windowPt.y);
                        windowPt.y = Math.min(cachedScreenBounds.y + cachedScreenBounds.height - 20, windowPt.y);

                    } else {
                        // For horizontal title bar, title pane must be fully visible
                        // on y-axis, and at least 20 pixel on x-axis.
                        windowPt.y = Math.max(cachedScreenBounds.y + titlePaneBounds.y, windowPt.y);
                        windowPt.y = Math.min(cachedScreenBounds.y + cachedScreenBounds.height -
                                titlePaneBounds.y - titlePaneBounds.height, windowPt.y);

                        windowPt.x = Math.max(cachedScreenBounds.x - windowSize.width + 20, windowPt.x);
                        windowPt.x = Math.min(cachedScreenBounds.x + cachedScreenBounds.width - 20, windowPt.x);
                    }
                }
                w.setLocation(windowPt);
            }
        } else if (dragCursor != 0) {
            Rectangle r = w.getBounds();
            Rectangle startBounds = new Rectangle(r);
            Dimension min = w.getMinimumSize();

            switch (dragCursor) {
                case Cursor.E_RESIZE_CURSOR:
                    adjust(r, min, 0, 0, pt.x + (dragWidth - dragOffsetX) -
                            r.width, 0);
                    break;
                case Cursor.S_RESIZE_CURSOR:
                    adjust(r, min, 0, 0, 0, pt.y + (dragHeight - dragOffsetY) -
                            r.height);
                    break;
                case Cursor.N_RESIZE_CURSOR:
                    adjust(r, min, 0, pt.y - dragOffsetY, 0,
                            -(pt.y - dragOffsetY));
                    break;
                case Cursor.W_RESIZE_CURSOR:
                    adjust(r, min, pt.x - dragOffsetX, 0,
                            -(pt.x - dragOffsetX), 0);
                    break;
                case Cursor.NE_RESIZE_CURSOR:
                    adjust(r, min, 0, pt.y - dragOffsetY,
                            pt.x + (dragWidth - dragOffsetX) - r.width,
                            -(pt.y - dragOffsetY));
                    break;
                case Cursor.SE_RESIZE_CURSOR:
                    adjust(r, min, 0, 0,
                            pt.x + (dragWidth - dragOffsetX) - r.width,
                            pt.y + (dragHeight - dragOffsetY) -
                            r.height);
                    break;
                case Cursor.NW_RESIZE_CURSOR:
                    adjust(r, min, pt.x - dragOffsetX,
                            pt.y - dragOffsetY,
                            -(pt.x - dragOffsetX),
                            -(pt.y - dragOffsetY));
                    break;
                case Cursor.SW_RESIZE_CURSOR:
                    adjust(r, min, pt.x - dragOffsetX, 0,
                            -(pt.x - dragOffsetX),
                            pt.y + (dragHeight - dragOffsetY) - r.height);
                    break;
                default:
                    break;
            }
            if (!r.equals(startBounds)) {
                w.setBounds(r);
                // Defer repaint/validate on mouseReleased unless dynamic
                // layout is active.
                if (Toolkit.getDefaultToolkit().isDynamicLayoutActive()) {
                    w.validate();
                    getRootPane().repaint();
                }
            }
        }
    }

    public void mouseEntered(MouseEvent ev) {
        Window w = (Window) ev.getSource();
        lastCursor = w.getCursor();
        mouseMoved(ev);
    }

    public void mouseExited(MouseEvent ev) {
        Window w = (Window) ev.getSource();
        w.setCursor(lastCursor);
    }

    public void mouseClicked(MouseEvent ev) {
        Window w = (Window) ev.getSource();
        Frame f = null;

        if (w instanceof Frame) {
            f = (Frame) w;
        } else {
            return;
        }

        Point convertedPoint = SwingUtilities.convertPoint(
                w, ev.getPoint(), getTitlePane());

        int state = f.getExtendedState();
        if (getTitlePane() != null &&
                getTitlePane().contains(convertedPoint)) {
            if ((ev.getClickCount() % 2) == 0 &&
                    ((ev.getModifiers() & InputEvent.BUTTON1_MASK) != 0)) {
                if (f.isResizable()) {
                    if ((state & Frame.MAXIMIZED_BOTH) != 0) {
                        f.setExtendedState(state & ~Frame.MAXIMIZED_BOTH);
                    } else {
                        f.setExtendedState(state | Frame.MAXIMIZED_BOTH);
                    }
                    return;
                }
            }
        }
    }

    Point getScreenLocation(MouseEvent e) {
        Point cursor = e.getPoint();
        Point target_location = this.target.getLocationOnScreen();
        return new Point(
            (int)(target_location.getX()+cursor.getX()),
            (int)(target_location.getY()+cursor.getY()));
    }
    /*public void mouseClicked(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {
        this.start_drag = this.getScreenLocation(e);
        this.start_loc = this.getWindow(this.target).getLocation();
        dragging=true;
    }

    public void mouseReleased(MouseEvent e) {
        dragging=false;
    }

    public void mouseDragged(MouseEvent e) {
        boolean quickDragg = (System.currentTimeMillis()-mouseDownTime)<=DOUBLE_CLICK_THRESHOLD;
        mouseDownTime = System.currentTimeMillis();
        if(dragging && SwingUtilities.isLeftMouseButton(e)) {
            Point current = this.getScreenLocation(e);
            Point offset = new Point(
                (int)current.getX()-(int)start_drag.getX(),
                (int)current.getY()-(int)start_drag.getY());
            JWindow frame = this.getWindow(target);
            Point new_location = new Point(
                (int)(this.start_loc.getX()+offset.getX()),
                (int)(this.start_loc.getY()+offset.getY()));
            frame.setLocation(new_location);
        }
    }
    public void mouseMoved(MouseEvent e) {}*/

    //ComponentListener methods
    public void componentHidden(ComponentEvent e) {}
    public void componentShown(ComponentEvent e) {}
    public void componentMoved(ComponentEvent e) {}
    public void componentResized(ComponentEvent e) {}
    /**
     * MouseInputHandler is responsible for handling resize/moving of
     * the Window. It sets the cursor directly on the Window when then
     * mouse moves over a hot spot.
     */
    /**
     * Returns the <code>JComponent</code> rendering the title pane. If this
     * returns null, it implies there is no need to render window decorations.
     *
     * @return the current window title pane, or null
     * @see #setTitlePane
     */
    private JComponent getTitlePane() {
        return titlePane;
    }

    /**
     * Returns the <code>JRootPane</code> we're providing the look and
     * feel for.
     */
    private JRootPane getRootPane() {
        return root;
    }
    private boolean isVertical(JRootPane root) {
        return root.getClientProperty("Quaqua.RootPane.isVertical") == Boolean.TRUE;
    }

}
